/* Yaml.g
//
// Copyright 2008 by Burt Harris
// Licensed under the Microsoft Permissive License (Ms-PL) V1.1
// See: http://www.codeplex.com/yaml/license
//
*/

grammar Yaml;

tokens {
	COMMA = ',';
	QUESTION_MARK = '?';
	INDENT;
	OUTDENT;
	}

@lexer::members {
		boolean inSingleQuoted, inDoubleQuoted, inBlockScalar;
		boolean bol = true;
		int flowNesting = 0;
		int indentLevel = 0;
		
		boolean BlockLevel() {
			return flowNesting == 0 && !(inSingleQuoted || inDoubleQuoted || inBlockScalar );
		}
		
		void trace( String s )
		{
			System.err.println( s 
				//+ " (" + state.tokenStartLine + ":" + (1+state.tokenStartCharPositionInLine) + ")" 
				);
		}
		
		boolean HasIndent()
		{
		  if (indentLevel == -1)
		  	return true;
			for (int i=0; i<indentLevel; i++)
			    if (input.LA(i+1) != ' ') return false;
			return true;
		}
		
		public void TraceLine() 
		{
			int i=0, ch;
			while ((ch = input.LA(i)) != -1)
			{
			  if (ch == '\n' || ch == '\r')
			  	break;
			  i++;
			}
			trace("");
			trace( "" + getLine() + "\t" + input.substring(input.index(),input.index()+i-2) );			
		}
		
		public Token nextToken()
		{
	  	state.token = null;
		  if (input.getCharPositionInLine() == 0)
		  {			
				state.channel = Token.DEFAULT_CHANNEL;
				state.tokenStartCharIndex = input.index();
				state.tokenStartCharPositionInLine = input.getCharPositionInLine();
				state.tokenStartLine = input.getLine();
				state.text = null;
				TraceLine();

				
				try {
					switch ( input.LA(1) ) {
					case '-':
						if (input.LA(2) != '-' || input.LA(3) != '-') break;
						mSTART_DOCUMENT();
						state.token = emit();
						break;
						
					case '.':
						if (input.LA(2) != '-' || input.LA(3) != '-') break;
						mEND_DOCUMENT();
						emit();
						break;
						
						
					case 'x':
						mDIRECTIVE();
						emit();
						break;
					}
				}
				catch (RecognitionException re) {
					reportError(re);
					// match() routine has already called recover()
				}	
				
				if (state.token == null)
				{
		  		if (HasIndent()) 
			  	{
			  		ConsumeIndent(); 
			  		if (indentLevel == -1 | (flowNesting == 0 && input.LA(1)==' '))
			  		{
			  			while (input.LA(1)==' ') input.consume();
			  			indentLevel = input.getCharPositionInLine();
			  			state.token = new CommonToken(INDENT); 
			  		}
			  	}
			  	else
			  	{
			  		indentLevel -= 2;
			  		state.token = new CommonToken(OUTDENT);
			  	}
			  }
		  }
		  if (state.token == null) state.token = super.nextToken();
		  if (state.token.getType() != -1) trace( YamlParser.tokenNames[state.token.getType()] + "\t" + state.token.toString() );
		  return state.token;
		}
		
		
		void ConsumeIndent()
		{
			for (int i=0; i<indentLevel; i++)
			{
				input.consume();
			}
		}
		
		boolean HasExtra()
		{
			boolean result = input.LA(1) == ' ';
		  return result;		    	
		}
		
		boolean AtCurrentIndent()
		{
		boolean result = state.tokenStartCharPositionInLine == indentLevel;
		return result; 
		}
	}
	
document
	:	( DIRECTIVE* START_DOCUMENT )?
		block_node
		END_DOCUMENT?
	;

LINE_BREAK 
	:	 {!(inSingleQuoted|inDoubleQuoted|inBlockScalar)}?=> 
		COMMENT? 
		NEWLINE 
		{if (flowNesting > 0) $channel = HIDDEN;}
		//BLANK_LINE*
	;

BLOCK_SEQUENCE_INDICATOR
	:	{BlockLevel()}?=> '-' (WS|NEWLINE)
	;

DIRECTIVE : {$pos==0}?=>'%' .* NEWLINE {indentLevel = -1;};
START_DOCUMENT : {$pos==0 /* START_DOCUMENT_TEST */ }?=> '---' {indentLevel = -1;};
END_DOCUMENT :	 {$pos==0}?=> '...' {indentLevel = -1;};
	
block_node
	: ((properties)=>properties)?
	  ( (implicit_map_entry)=> block_mapping
	  | alias_node
		| flow_content 
		| block_sequence
		| block_mapping
		| LITERAL
		| FOLDED
		)
	;
	
block_sequence 
	: BLOCK_SEQUENCE_INDICATOR block_node
		(BLOCK_SEQUENCE_INDICATOR block_node)* 
		OUTDENT
	;
	
block_mapping 
  options { backtrack=true; }
	:	block_map_entry
		block_map_entry+
		OUTDENT
	;

block_map_entry
  options { backtrack=true; }
	:	'?' block_indented (COLON block_indented)?
	| implicit_map_entry
	;

implicit_map_entry
	:	( ( PLAIN
			| SINGLE_QUOTED_SCALAR
			| DOUBLE_QUOTED_SCALAR
			)? 
			( COLON block_node? 
			)
		)
	;

	
block_indented 
	: INDENT block_node OUTDENT
	| block_node
	;


flow_node 
  : alias_node
  | properties flow_content?
  | flow_content
  ;
  
alias_node
	:	ALIAS
	;
      
properties 
 	:	LABEL TAG? 
 	| TAG LABEL?
 	;
  
flow_collection 
	:	OPEN_BRACKET ((flow_item ( ',' flow_item )* ','?))? CLOSE_BRACKET
  | OPEN_BRACE ((flow_item (',' flow_item)* ','?))? CLOSE_BRACE
  ;
  
flow_scalar
  : 
  | SINGLE_QUOTED_SCALAR
  | DOUBLE_QUOTED_SCALAR
  | PLAIN (PLAIN)*
  ;
  
flow_content 
	:	flow_collection
	| flow_scalar
	;

flow_item 
 	: flow_node 
	| '?' flow_node? ':' flow_node?  
  | ':' flow_node? 
 	| DOUBLE_QUOTED_SCALAR ':' flow_node
 	| plain_key ':' flow_node
 	;
 
plain_key : PLAIN;
	         	
fragment DIGIT :	'0'..'9' ;
fragment HEX_DIGIT	: '0' .. '9' | 'A'..'F' | 'a'..'f';
fragment HEX_BYTE	: HEX_DIGIT HEX_DIGIT;
fragment HEX_16			:	HEX_BYTE HEX_BYTE;
fragment HEX_32			: HEX_16 HEX_16;
    
fragment WORD_CHAR :	'A'..'Z' | 'a'..'z'| '0' .. '9' | '-';
    	
fragment URI_CHAR
	:	WORD_CHAR
  | '%' HEX_DIGIT HEX_DIGIT
  | ';' | '/' | '?' | ':' | '@' | '&' | '=' | '+' | '$' | ','
	| '_' | '.' | '!' | '~' | '*' | '\'' | '(' | ')' | '[' | ']' 
	;
			
fragment TAG_CHAR	// Characters that can make up a tag handle
	:	WORD_CHAR
  | '%' HEX_DIGIT HEX_DIGIT
  | ';' | '/' | '?' | ':' | '@' | '&' | '=' | '+' | '$' 
	| '_' | '.' | '~' | '*' | '\'' | '(' | ')'  
	;
			
fragment ANCHOR_CHAR 
	:	'!'..'+' | '-'..'Z' | '\\' | '^'..'z' | '|' | '~'
	| '\u00A0'..'\uD7FF'
	| '\uD800'..'\uDBFF' '\uDC00'..'\uDFFF'		// UTF-16 surogate pair
	;
		
fragment ESCAPED_CHAR
	: '\\'
		(	'0' 
		| 'a'
		| 'b'
		| 't'
		| '\t'
		| 'n'
		| 'v'
		| 'r'
		| 'f'
		| 'e'
		| ' '
		| '"' 
		| '/' 						/* ESCAPE SLASH requred for JSON compatability */
		| '\\'
		| 'N'
		| '_'
		| 'L'
		| 'P'
		| 'x' HEX_BYTE
		| 'u' HEX_16
		| 'U' HEX_32
		)
	; 
			
			
fragment DOUBLE_QUOTED_CHAR
	: '\t' | ' ' 
	| ('!' | '#'..'[' | ']'..'~' 
	| '\u00A0'..'\uD7FF'
	| '\uD800'..'\uDBFF' '\uDC00'..'\uDFFF'
	| ESCAPED_CHAR) 
	;
		
DOUBLE_QUOTED_SCALAR 
	: '"'  {inDoubleQuoted = true; }
		DOUBLE_QUOTED_CHAR+ 
		'"'  {inDoubleQuoted = false; }
	;
		
SINGLE_QUOTED_SCALAR 
	: '\'' {inSingleQuoted = true;} 
		( '\t' 
		| ' ' .. '&' | '(' .. '~' // Printable ASCII - SINGLE_QUOTE
		| ('\u00A0'..'\uD7FF'
		| ('\uD800'..'\uDBFF') ('\uDC00'..'\uDFFF'))
		| '\'' '\''
		)+ 
		'\''  {inSingleQuoted = false;}
	;
		
OPEN_BRACKET :		'[' { flowNesting++; } ;	
CLOSE_BRACKET :	']' { flowNesting--; } ;	
OPEN_BRACE :			'{' { flowNesting++; } ;		
CLOSE_BRACE :		'}' { flowNesting--; } ;	
fragment NEWLINE 
	:	'\r' '\n'?
	| '\n'
	| '\u0085'
	;  
            
     
ALIAS :	 '*' (ANCHOR_CHAR+) ;
LABEL :	 '&' (ANCHOR_CHAR+) ;

TAG 
	:	'!'
		( '<' URI_CHAR+ '>'
		| TAG_CHAR+
		| '!' TAG_CHAR+
		| WORD_CHAR+ '!' TAG_CHAR+
		)
		;
		
fragment WS :	 (' ' | '\t' )+ ;

IGNORED_WS :	 WS {$channel = HIDDEN;};

fragment COMMENT 
	:	'#' 
		( ' '..'~' 
		| '\u00A0'..'\uD7FF'
		| '\uD800'..'\uDBFF' '\uDC00'..'\uDFFF'	// UTF-16 surrogate pairs
		)*
	;
	
COLON
	:	':' 
	  { trace("COLON");
	  	if (flowNesting == 0) 
	  	{
	  		trace("BLOCK COLON");
	  		// indentLevel += 2; 
	  	}
	  }
	;

fragment BLANK_LINE
	:	((' '|'\t')* COMMENT? NEWLINE)	// Gated match
	=> (' '|'\t')* COMMENT? NEWLINE 	// Consume it
	;
	
PLAIN 
	:	PLAIN_FIRST
		( (PLAIN_SAFE | '#')
		| ((' '|'\t')+ PLAIN_SAFE)=>(' '|'\t')+ PLAIN_SAFE
		| ((':')+ PLAIN_SAFE)=>(':')+ PLAIN_SAFE
		)* 
	; 

fragment PLAIN_FIRST 
	:	'$' | '(' | ')' | '+' | ';' | '<' | '~' 
	| '.' .. '9' | 'A'..'Z' | 'a'..'z'  | '\\' 
	| '\u00A0'..'\uD7FF'
	| '\uD800'..'\uDBFF' '\uDC00'..'\uDFFF'	// UTF-16 surrogate pairs
	| ( '?' | ':' | '-') PLAIN_SAFE 
	;

fragment PLAIN_SAFE 
	:	'!' .. '"' | '$' ..'+' 
	| '-' ..'9'
	| ';' .. 'Z' | '\\' 
	| '^' .. 'z' 	| '|' | '~' 
	| '\u00A0'..'\uD7FF'
	| '\uD800'..'\uDBFF' '\uDC00'..'\uDFFF'	// UTF-16 surrogate pairs
	|	({ flowNesting == 0 }?=>(',' | '[' | ']' | '{' | '}'))       	// Characters allows in block context
	;
	
fragment NONSPACE 
	:	 '!'..'~' // | NON_ASCII_PRINTABLE 
	;
	

LITERAL 
	:	'|' 
		BLOCK_SCALAR_HEADER 
		{ inBlockScalar = true; }
			( BLOCK_SCALAR_LINE )*	
		{ inBlockScalar = false; }
	;
	

FOLDED 
	:	'>' 
		BLOCK_SCALAR_HEADER 
		{ inBlockScalar = true; }
			( BLOCK_SCALAR_LINE 
			)*	
	;	

fragment INDENTATION_INDICATOR
	:	DIGIT			
	| 					{indentLevel+= 2;}
	;
	
fragment CHOMPING
	:	'-'		
	| '+'
	|
	;

fragment BLOCK_SCALAR_HEADER 
	:	CHOMPING INDENTATION_INDICATOR WS? ('#' .*)? NEWLINE
	;
	
fragment BLOCK_SCALAR_LINE 
  :	 {HasIndent()}?=> {ConsumeIndent();} .* NEWLINE
  ;

// INDENT 	:	{$pos==0}?=> ({($pos < indentlevel)}? ' ' )* ;
